import json
import copy
import traceback
import configparser
import requests
import yaml
from infoworks.core.iw_authentication import get_bearer_token
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.url_builder import get_pipeline_group_base_url,list_domains_url,list_pipelines_url
from infoworks.sdk.cicd.upload_configurations.update_configurations import InfoworksDynamicAccessNestedDict
from infoworks.sdk.cicd.upload_configurations.local_configurations import PRE_DEFINED_MAPPINGS

class PipelineGroup:
    def __init__(self, pipeline_group_config_path, environment_id, storage_id, interactive_id,
                 replace_words="", secrets=None):
        self.storage_id = storage_id
        self.interactive_id = interactive_id
        self.environment_id = environment_id
        self.secrets = secrets
        with open(pipeline_group_config_path, 'r') as file:
            json_string = file.read()
            if replace_words != "":
                for key, value in [item.split("->") for item in replace_words.split(";")]:
                    json_string = json_string.replace(key, value)
        self.configuration_obj = IWUtils.ejson_deserialize(json_string)

    def update_mappings_for_configurations(self, mappings):
        config = configparser.ConfigParser()
        config.read_dict(mappings)
        d = InfoworksDynamicAccessNestedDict(self.configuration_obj)
        for section in config.sections():
            if section in PRE_DEFINED_MAPPINGS:
                continue
            print("section:", section)
            try:
                final = d.setval(section.split("$"), dict(config.items(section)))
                print(f"section replacement:{d.getval(section.split('$'))}")
            except KeyError as e:
                pass
        self.configuration_obj = d.data


    def list_pipelines(self, pipeline_group_obj,domain_id=None, params=None):
        """
        Function to list the pipelines
        :param domain_id: Entity identified for domain
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response list
        """

        if None in {domain_id}:
            pipeline_group_obj.logger.error("Domain ID cannot be None")
            raise Exception("Domain ID cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipelines = list_pipelines_url(pipeline_group_obj.client_config, domain_id) \
                                + IWUtils.get_query_params_string_from_dict(params=params)

        pipelines_list = []
        headers={'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}
        try:
            response = requests.request("GET", url_to_list_pipelines,
                                 headers=headers, verify=False)
            if response.status_code == 406:
                headers = pipeline_group_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", url_to_list_pipelines, headers=headers, verify=False)
            if response.status_code==200:
                response = response.json()
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipelines_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=pipeline_group_obj.client_config['ip'],
                                                                      port=pipeline_group_obj.client_config['port'],
                                                                      protocol=pipeline_group_obj.client_config['protocol'],
                                                                      )
                    response = requests.request("GET", nextUrl, headers=headers,verify=False)
                    if response.status_code == 406:
                        headers = pipeline_group_obj.regenerate_bearer_token_if_needed(
                            {'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                             'Content-Type': 'application/json'})
                        response = requests.request("GET", nextUrl, headers=headers, verify=False)
                    response = response.json()
                    result = response.get("result", None)
                    if result is None:
                        return response
                response["result"] = pipelines_list
                response["message"] = initial_msg
            return response
        except Exception as e:
            pipeline_group_obj.logger.error("Error in listing pipelines")
            raise Exception("Error in listing pipelines" + str(e))

    def create(self, pipeline_group_obj, domain_id, domain_name):

        pipeline_group_name = self.configuration_obj["name"]
        pipeline_group_id = None
        final_domain_id=domain_id
        if self.environment_id is not None:
            self.configuration_obj["environment_id"]=self.environment_id
        if domain_name is not None and domain_id is None:
            domains_url_base = list_domains_url(pipeline_group_obj.client_config)
            filter_condition = IWUtils.ejson_serialize({"name": domain_name})
            domains_url = domains_url_base + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
            response = requests.request("GET", domains_url, headers={
                'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, verify=False)
            if response.status_code == 406:
                headers = pipeline_group_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", domains_url, headers=headers, verify=False)
            if response is not None:
                result = response.json().get("result", [])
                if len(result) > 0:
                    final_domain_id = result[0]["id"]
                    self.configuration_obj["domain_id"] = final_domain_id
        pipeline_group_details_url = get_pipeline_group_base_url(pipeline_group_obj.client_config,final_domain_id)+"?filter={\"name\":\""+pipeline_group_name+"\"}"
        headers = pipeline_group_obj.regenerate_bearer_token_if_needed(
            {'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
             'Content-Type': 'application/json'})
        pipeline_group_details_response = requests.request("GET", pipeline_group_details_url, headers=headers, verify=False)
        pipeline_group_details_parsed_response = IWUtils.ejson_deserialize(
            pipeline_group_details_response.content)
        pipeline_name_lookup={}
        list_pipelinesparsed_response = self.list_pipelines(pipeline_group_obj=pipeline_group_obj,domain_id=final_domain_id)
        for item in list_pipelinesparsed_response["result"]:
            pipeline_name_lookup[item["name"]]=item["id"]
        for pipeline in self.configuration_obj["pipelines"]:
            pipeline["pipeline_id"]=pipeline_name_lookup[pipeline["name"]]
        if len(pipeline_group_details_parsed_response["result"])>0:
            pipeline_group_obj.logger.info(
                'Pipeline group already exists. {} {}')
            pipeline_group_obj.logger.info('Getting the existing pipeline group Id with given name.')
            pipeline_group_id = pipeline_group_details_parsed_response["result"][0]["pipeline_group_details"]["id"]
            del self.configuration_obj["id"]
            configuration_obj = copy.deepcopy(self.configuration_obj)
            for item in self.configuration_obj:
                if item in ["createdAt","createdBy","modifiedAt","modifiedBy","environment_configurations"]:
                    del configuration_obj[item]
            self.configuration_obj = configuration_obj
            for pipeline in self.configuration_obj["pipelines"]:
                if pipeline.get("name",None) is not None:
                    del pipeline["name"]

            pipeline_group_create_url = get_pipeline_group_base_url(pipeline_group_obj.client_config, final_domain_id)+pipeline_group_id
            response = pipeline_group_obj.call_api("PUT", pipeline_group_create_url, {
                'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, data=self.configuration_obj)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            print("parsed_response", parsed_response)
            if response.status_code == 200 and "result" in parsed_response:
                pipeline_group_obj.logger.info("Successfully updated the pipeline group")
                print("Successfully updated the pipeline group")
                pipeline_group_id = parsed_response["result"]["id"]
        else:
            del self.configuration_obj["id"]
            configuration_obj = copy.deepcopy(self.configuration_obj)
            for item in self.configuration_obj:
                if item in ["createdAt", "createdBy", "modifiedAt", "modifiedBy", "environment_configurations"]:
                    del configuration_obj[item]
            self.configuration_obj = configuration_obj
            for pipeline in self.configuration_obj["pipelines"]:
                if pipeline.get("name", None) is not None:
                    del pipeline["name"]

            pipeline_group_create_url = get_pipeline_group_base_url(pipeline_group_obj.client_config,final_domain_id)
            response = pipeline_group_obj.call_api("POST", pipeline_group_create_url, {
                'Authorization': 'Bearer ' + pipeline_group_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, data=self.configuration_obj)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            print("parsed_response",parsed_response)
            if response.status_code == 200 and "result" in parsed_response:
                pipeline_group_obj.logger.info("Successfully created and configured the pipeline group")
                print("Successfully created and configured the pipeline group")
                pipeline_group_id=parsed_response["result"]["id"]
        return pipeline_group_id, self.configuration_obj["domain_id"]




import traceback

from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.cicd.upload_configurations.pipelines import Pipeline
from infoworks.sdk.cicd.upload_configurations.pipeline_group import PipelineGroup
from pathlib import Path
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.local_configurations import Response
import os.path
import queue
import threading
import networkx as nx


class WrapperPipeline(BaseClient):
    def __init__(self):
        super().__init__()

    def __wrapper_get_environment_details(self, environment_id=None, params=None):
        if params is None and environment_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_environments = url_builder.get_environment_details(
            self.client_config, environment_id) + IWUtils.get_query_params_string_from_dict(params=params)
        env_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_environments,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if environment_id is not None:
                    env_details.extend(result)
                else:
                    while len(result) > 0:
                        env_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=env_details)
        except Exception as e:
            self.logger.error("Error in getting environment details")

    def __wrapper_get_storage_details(self, environment_id, storage_id=None, params=None):
        if params is None and storage_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_storages = url_builder.get_environment_storage_details(
            self.client_config, environment_id, storage_id) + IWUtils.get_query_params_string_from_dict(params=params)
        storage_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_storages,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if storage_id is not None:
                    storage_details.extend(result)
                else:
                    while len(result) > 0:
                        storage_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=storage_details)
        except Exception as e:
            self.logger.error("Error in getting storage details")

    def __wrapper_get_compute_template_details(self, environment_id, compute_id=None, is_interactive=False,
                                               params=None):
        if params is None and compute_id is None:
            params = {"limit": 20, "offset": 0}
        if is_interactive:
            url_to_list_computes = url_builder.get_environment_interactive_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        else:
            url_to_list_computes = url_builder.get_environment_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        compute_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_computes,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                if compute_id is not None:
                    compute_details.extend(result)
                else:
                    while len(result) > 0:
                        compute_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=compute_details)
        except Exception as e:
            self.logger.error("Error in getting compute template details")
    def cicd_create_configure_pipeline(self, configuration_file_path, domain_id=None, domain_name=None,
                                       override_configuration_file=None,
                                       replace_words="", read_passwords_from_secrets=False, env_tag="", secret_type=""):
        """
        Function to create and configure pipeline using the pipeline configuration JSON file
        Pass either domain_id or domain_name.If both are not passed the name of the domain should be first part of the file name
        :param configuration_file_path: Path of the file with pipeline configurations to be imported
        :param domain_id: Domain id to which the pipeline belongs to
        :param domain_name: Domain name to which the pipeline belongs to
        :param override_configuration_file: Path of the file with override keys for dataconnection properties
        :param replace_words: Pass the strings to be replaced in the configuration file. Example: DEV->PROD;dev->prod
        :param read_passwords_from_secrets: True/False. If True all the pipeline related passwords are read from encrypted file name passed
        """
        try:
            if domain_id is None and domain_name is None:
                domain_name = Path(configuration_file_path).name.split("#")[0]
                domain_mappings = self.mappings.get("domain_name_mappings", {})
                if domain_mappings != {}:
                    if domain_name.lower() in domain_mappings.keys():
                        domain_name = domain_mappings.get(domain_name.lower(), domain_name)
            env_id = self.client_config.get("default_environment_id", None)
            storage_id = self.client_config.get("default_storage_id", None)
            compute_id = self.client_config.get("default_compute_id", None)
            with open(configuration_file_path, 'r') as file:
                json_string = file.read()
            configuration_obj = IWUtils.ejson_deserialize(json_string)
            environment_configurations = configuration_obj["environment_configurations"]
            if env_id is None and "environment_mappings" in self.mappings:
                env_name = self.mappings["environment_mappings"].get(environment_configurations["environment_name"],
                                                                     environment_configurations["environment_name"])
                if env_name is not None:
                    result = self.__wrapper_get_environment_details(params={"filter": {"name": env_name}})
                    env_id = result["result"]["response"][0]["id"] if len(result["result"]["response"]) > 0 else None
            if storage_id is None and "storage_mappings" in self.mappings:
                storage_name = self.mappings["storage_mappings"].get(
                    environment_configurations["environment_storage_name"],
                    environment_configurations["environment_storage_name"])
                if storage_name is not None:
                    result = self.__wrapper_get_storage_details(environment_id=env_id,
                                                                params={"filter": {"name": storage_name}})
                    storage_id = result["result"]["response"][0]["id"] if len(
                        result["result"]["response"]) > 0 else None
            if compute_id is None and "compute_mappings" in self.mappings:
                compute_name = self.mappings["compute_mappings"].get(
                    environment_configurations["environment_compute_template_name"],
                    environment_configurations["environment_compute_template_name"])
                if compute_name is not None:
                    result = self.__wrapper_get_compute_template_details(environment_id=env_id,
                                                                         params={"filter": {"name": compute_name}})
                    if len(result["result"]["response"]) != 0:
                        compute_id = result["result"]["response"][0]["id"]
                    else:
                        result = self.__wrapper_get_compute_template_details(environment_id=env_id, is_interactive=True,
                                                                             params={"filter": {"name": compute_name}})
                        if len(result["result"]["response"]) != 0:
                            compute_id = result["result"]["response"][0]["id"]
            if env_id is None:
                print("No env id and no mapping found")
                raise Exception("No env id and no mapping found")
            pl_obj = Pipeline(configuration_file_path, env_id, storage_id, compute_id, replace_words,
                              self.secrets_config)
            pl_obj.configuration_obj = configuration_obj
            pl_obj.update_mappings_for_configurations(self.mappings)
            pipeline_id, domain_id = pl_obj.create(self, domain_id, domain_name)

            if pipeline_id is not None:
                status = pl_obj.configure(self, pipeline_id, domain_id, override_configuration_file, self.mappings,
                                              read_passwords_from_secrets, env_tag=env_tag, secret_type=secret_type)
        except Exception as e:
            self.logger.error(str(e))
            print(str(e))
            self.logger.error(traceback.format_exc())
            print(traceback.format_exc())

    def cicd_create_configure_pipeline_group(self, configuration_file_path, domain_id=None, domain_name=None,
                                             override_configuration_file=None,
                                             replace_words="", read_passwords_from_secrets=False, env_tag="",
                                             secret_type=""):
        """
        Function to create and configure pipeline group using the pipeline group configuration JSON file
        Pass either domain_id or domain_name.If both are not passed the name of the domain should be first part of the file name
        :param configuration_file_path: Path of the file with pipeline configurations to be imported
        :param domain_id: Domain id to which the pipeline belongs to
        :param domain_name: Domain name to which the pipeline belongs to
        :param override_configuration_file: Path of the file with override keys for dataconnection properties
        :param replace_words: Pass the strings to be replaced in the configuration file. Example: DEV->PROD;dev->prod
        :param read_passwords_from_secrets: True/False. If True all the pipeline related passwords are read from encrypted file name passed
        """
        try:
            if domain_id is None and domain_name is None:
                domain_name = Path(configuration_file_path).name.split("#")[0]
                domain_mappings = self.mappings.get("domain_name_mappings", {})
                if domain_mappings != {}:
                    if domain_name.lower() in domain_mappings.keys():
                        domain_name = domain_mappings.get(domain_name.lower(), domain_name)
            env_id = self.client_config.get("default_environment_id", None)
            storage_id = self.client_config.get("default_storage_id", None)
            compute_id = self.client_config.get("default_compute_id", None)
            with open(configuration_file_path, 'r') as file:
                json_string = file.read()
            configuration_obj = IWUtils.ejson_deserialize(json_string)
            environment_configurations = configuration_obj["environment_configurations"]
            if env_id is None and "environment_mappings" in self.mappings:
                env_name = self.mappings["environment_mappings"].get(environment_configurations["environment_name"],
                                                                     environment_configurations["environment_name"])
                if env_name is not None:
                    result = self.__wrapper_get_environment_details(params={"filter": {"name": env_name}})
                    env_id = result["result"]["response"][0]["id"] if len(result["result"]["response"]) > 0 else None
            if storage_id is None and "storage_mappings" in self.mappings:
                storage_name = self.mappings["storage_mappings"].get(
                    environment_configurations["environment_storage_name"],
                    environment_configurations["environment_storage_name"])
                if storage_name is not None:
                    result = self.__wrapper_get_storage_details(environment_id=env_id,
                                                                params={"filter": {"name": storage_name}})
                    storage_id = result["result"]["response"][0]["id"] if len(
                        result["result"]["response"]) > 0 else None
            if compute_id is None and "compute_mappings" in self.mappings:
                compute_name = self.mappings["compute_mappings"].get(
                    environment_configurations["environment_compute_template_name"],
                    environment_configurations["environment_compute_template_name"])
                if compute_name is not None:
                    result = self.__wrapper_get_compute_template_details(environment_id=env_id,
                                                                         params={"filter": {"name": compute_name}})
                    if len(result["result"]["response"]) != 0:
                        compute_id = result["result"]["response"][0]["id"]
                    else:
                        result = self.__wrapper_get_compute_template_details(environment_id=env_id, is_interactive=True,
                                                                             params={"filter": {"name": compute_name}})
                        if len(result["result"]["response"]) != 0:
                            compute_id = result["result"]["response"][0]["id"]
            if env_id is None:
                print("No env id and no mapping found")
                raise Exception("No env id and no mapping found")
            pl_grp_obj = PipelineGroup(configuration_file_path, env_id, storage_id, compute_id, replace_words,
                                       self.secrets_config)
            pl_grp_obj.configuration_obj = configuration_obj
            pl_grp_obj.update_mappings_for_configurations(self.mappings)
            pipeline_group_id, domain_id = pl_grp_obj.create(self, domain_id, domain_name)
        except Exception as e:
            self.logger.error(str(e))
            print(str(e))
            self.logger.error(traceback.format_exc())
            print(traceback.format_exc())

    def __execute(self, thread_number, q):
        while True:
            try:
                print('%s: Looking for the next task ' % thread_number)
                task = q.get()
                print(f'\nThread Number {thread_number} processed {task}')
                entity_type = task["entity_type"]
                if entity_type == "pipeline":
                    replace_words = task["replace_words"] if task["replace_words"] else ""
                    override_configuration_file = ""
                    read_passwords_from_secrets = task.get("read_passwords_from_secrets", False)
                    self.cicd_create_configure_pipeline(task["pipeline_config_path"], None,
                                                        task["domain_name"],
                                                        override_configuration_file, replace_words,
                                                        read_passwords_from_secrets)
            except Exception as e:
                print(str(e))
            q.task_done()

    def cicd_create_pipelineartifacts_in_bulk(self, maintain_lineage, pipeline_gexf_lineage_file=None,
                                              configurations_base_path=None,
                                              pl_replace_words=None,
                                              override_configuration_file=None, read_passwords_from_secrets=False):
        """
        Function to configure pipelines in bulk
        :param maintain_lineage: True/False. If True pass the pipeline_gexf_lineage_file which has a dag of pipeline dependencies
        :param pipeline_gexf_lineage_file: File with pipeline dependencies
        :param configurations_base_path: Path with all the source configuration dumps
        :param pl_replace_words: Pass the strings to be replaced in the configuration file. Example: DEV->PROD;dev->prod
        :param override_configuration_file: Path of the file with override keys for dataconnection properties
        :param read_passwords_from_secrets: True/False. If True all the source related passwords are read from encrypted file name passed
        """
        num_fetch_threads = 10
        job_queue = queue.Queue(maxsize=100)
        for i in range(num_fetch_threads):
            worker = threading.Thread(target=self.__execute, args=(i, job_queue,))
            worker.setDaemon(True)
            worker.start()

        if maintain_lineage:
            if pipeline_gexf_lineage_file:
                path = os.path.join(configurations_base_path, "modified_files/pipeline.gexf")
            else:
                path = pipeline_gexf_lineage_file
            graph = nx.read_gexf(path)
            for item in self._topological_sort_grouping(graph):
                for i in item:
                    if i != "root":
                        pipeline_file = graph._node[i]['filename'].strip()
                        domain_name = pipeline_file.split("#")[0]
                        domain_mappings = self.mappings.get("domain_name_mappings", {})
                        if domain_mappings != {}:
                            if domain_name.lower() in domain_mappings.keys():
                                domain_name = domain_mappings.get(domain_name.lower(), domain_name)
                        pipeline_args = {"entity_type": "pipeline",
                                         "pipeline_config_path": os.path.join(configurations_base_path, "pipeline",
                                                                              pipeline_file.strip()),
                                         "domain_name": domain_name, "replace_words": pl_replace_words,
                                         "override_configuration_file": override_configuration_file,
                                         "read_passwords_from_secrets": read_passwords_from_secrets}
                        job_queue.put(pipeline_args)
                print('*** Main thread waiting to complete all pending pipeline creation/configuration requests ***')
                job_queue.join()
                print('*** Done with All Tasks ***')
        else:
            with open(os.path.join(configurations_base_path, "modified_files", "pipeline.csv"),
                      "r") as pipeline_files_fp:
                for pipeline_file in pipeline_files_fp.readlines():
                    pl_name = pipeline_file.strip()
                    domain_name = pl_name.split("#")[0]
                    domain_mappings = self.mappings.get("domain_name_mappings", {})
                    if domain_mappings != {}:
                        if domain_name.lower() in domain_mappings.keys():
                            domain_name = domain_mappings.get(domain_name.lower(), domain_name)
                    pipeline_args = {"entity_type": "pipeline",
                                     "pipeline_config_path": os.path.join(configurations_base_path, "pipeline",
                                                                          pipeline_file.strip()),
                                     "domain_name": domain_name, "replace_words": pl_replace_words,
                                     "override_configuration_file": override_configuration_file}
                    job_queue.put(pipeline_args)

            print('*** Main thread waiting to complete all pipeline configuration requests ***')
            job_queue.join()
            print('*** Done with Pipeline Configurations ***')






def get_bearer_token_url(config):
    """
    returns URL to get bearer token  using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to get the bearer_token
    """
    request = '{protocol}://{ip}:{port}/v3/security/token/access'.format(ip=config['ip'], port=config['port'],
                                                                         protocol=config['protocol'])
    return request


def validate_bearer_token_url(config):
    """
    returns URL to validate bearer token using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to validate bearer token
    """
    request = '{protocol}://{ip}:{port}/v3/security/token/validate'.format(ip=config['ip'], port=config['port'],
                                                                           protocol=config['protocol'])
    return request


def list_domains_url(config):
    """
    returns URL to list domains using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_pipelines_url(config, domain_id):
    """
    returns URL to list pipelines under given domain using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list pipelines
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines'.format(ip=config['ip'], port=config['port'],
                                                                                 protocol=config['protocol'],
                                                                                 domain_id=domain_id)
    return request


def list_sources_url(config):
    """
    returns URL to list sources using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list sources
    """
    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_tables_under_source(config, source_id):
    """
    returns URL to list all tables in source using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity identifier
    :type source_id: string
    :return: url to list all tables in source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables'.format(ip=config['ip'], port=config['port'],
                                                                              protocol=config['protocol'],
                                                                              source_id=source_id)
    return request


def get_general_configs_url(config):
    """
    returns URL to create source using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to get admin configurations
    """
    request = '{protocol}://{ip}:{port}/v3/admin/configuration/general'.format(ip=config['ip'], port=config['port'],
                                                                               protocol=config['protocol'])
    return request


def get_source_details_url(config):
    """
    returns URL to list sources using v3 api
    :param config: client configurations
    :type config: dict
    :param filter: filter criteria
    :type: filter: string
    :return: source details matching filter criteria
    """

    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])

    return request


def source_info(config, source_id):
    """
    returns URL to get source info using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source id
    :type: source_id: string
    :return: URL to get source information
    """

    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}'.format(ip=config['ip'], port=config['port'],
                                                                       protocol=config['protocol'], source_id=source_id)

    return request


def create_source_url(config):
    """
    returns URL to create source using v3 api
    :param config: client configurations
    :type config: dict
    :return: url for the source creation
    """
    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def get_source_configurations_url(config, source_id):
    """
    returns URL to get source configuration using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to get source configuration
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations'.format(ip=config['ip'],
                                                                                      port=config['port'],
                                                                                      protocol=config['protocol'],
                                                                                      source_id=source_id)
    return request


def get_source_connection_details_url(config, source_id):
    """
    returns URL to get or update source connection details using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to get/update source connection details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations/connection'.format(ip=config['ip'],
                                                                                                 port=config['port'],
                                                                                                 protocol=config[
                                                                                                     'protocol'],
                                                                                                 source_id=source_id)
    return request


def interactive_job_poll_url(config, source_id, job_id):
    """
    returns URL to import poll interactive jobs using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param job_id: job id of interactive job
    :type config: string
    :return: url to poll interactive job
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/interactive-jobs/{interactive_job_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        interactive_job_id=job_id)

    return request


def browse_source_tables_url(config, source_id):
    """
    returns URL to browse source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to browse source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/source_tables'.format(ip=config['ip'],
                                                                                     port=config['port'],
                                                                                     protocol=config['protocol'],
                                                                                     source_id=source_id)
    return request


def add_tables_to_source_url(config, source_id):
    """
    returns URL to browse source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to browse source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/source_tables'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            protocol=config['protocol'],
                                                                                            source_id=source_id)
    return request

def add_query_tables_to_source_url(config, source_id):
    """
    returns URL to add query tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to browse source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/query_tables'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            protocol=config['protocol'],
                                                                                            source_id=source_id)
    return request

def configure_tables_and_tablegroups_url(config, source_id):
    """
    returns URL to configure source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to configure source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/config-migration'.format(ip=config['ip'],
                                                                                        port=config['port'],
                                                                                        protocol=config['protocol'],
                                                                                        source_id=source_id)
    return request


def create_table_group_url(config, source_id):
    """
        returns URL to create table groups using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: source entity id
        :type source_id: string
        :return: url to create table groups using v3 api
        """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups'.format(ip=config['ip'],
                                                                                    port=config['port'],
                                                                                    protocol=config['protocol'],
                                                                                    source_id=source_id)
    return request


def get_compute_template_details_url(config, environment_id):
    """
    returns URL to get compute template details using v3 api
    :param config: client configurations
    :type config: dict
    :param environment_id: environment id
    :type environment_id: string
    :return: url to get compute template details
    """
    request = '{protocol}://{ip}:{port}/v3/admin/environment/{environment_id}/environment-interactive-clusters'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        environment_id=environment_id)
    return request


def get_test_connection_url(config, source_id):
    """
    returns URL to test rdbms source connection using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to test rdbms source connection
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs'.format(ip=config['ip'], port=config['port'],
                                                                            protocol=config['protocol'],
                                                                            source_id=source_id)
    return request


def get_job_status_url(config, job_id=None):
    """
    returns URL for checking job progress

    :param config: client configurations
    :type config: dict
    :param job_id: Identifier for Job
    :type job_id: string
    :return: url for the job status
    """
    if job_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/jobs'.format(ip=config['ip'],
                                                                  port=config['port'],
                                                                  protocol=config['protocol'])
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}'.format(ip=config['ip'],
                                                                           port=config['port'],
                                                                           job_id=job_id,
                                                                           protocol=config['protocol'])
    return request


def configure_source_url(config, source_id):
    """
    returns URL to configure the source using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to configure the source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/config-migration'.format(ip=config['ip'],
                                                                                        port=config['port'],
                                                                                        protocol=config['protocol'],
                                                                                        source_id=source_id)
    return request


def create_domain_url(config):
    """
    returns URL for creating domain

    :param config: client configurations
    :type config: dict
    :return: url for the domain creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains'.format(ip=config['ip'],
                                                           port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_users_url(config):
    """
    returns URL to list users using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list users
    """
    request = '{protocol}://{ip}:{port}/v3/admin/users'.format(ip=config['ip'], port=config['port'],
                                                               protocol=config['protocol'])
    return request


def add_user_to_domain_url(config, user_id):
    """
    returns URL to add user to domains

    :param config: client configurations
    :type config: dict
    :param user_id: Identifier for user
    :type user_id: string
    :return: url for the add user to domain
    """
    request = '{protocol}://{ip}:{port}/v3/admin/users/{user_id}/accessible-domains'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            user_id=user_id,
                                                                                            protocol=config['protocol'])
    return request


def add_sources_to_domain_url(config, domain_id):
    """
    returns URL to add sources to domains

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url to add sources to domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-sources'.format(ip=config['ip'],
                                                                                          port=config['port'],
                                                                                          domain_id=domain_id,
                                                                                          protocol=config['protocol'])
    return request


def create_pipeline_url(config, domain_id):
    """
    returns URL to create pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url for the pipeline creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines'.format(ip=config['ip'], port=config['port'],
                                                                                 protocol=config['protocol'],
                                                                                 domain_id=domain_id)
    return request


def get_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to get pipeline details using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def delete_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to delete pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to delete pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def update_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to update pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to update pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def configure_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to configure pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to configure pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/config-migration'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def list_pipeline_versions_url(config, domain_id, pipeline_id):
    """
    returns URL to list pipeline versions using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to list pipeline versions
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def make_pipeline_version_active_url(config, domain_id, pipeline_id, version_id):
    """
    returns URL to make pipeline version active
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :param version_id: Identifier for pipeline
    :type version_id: string
    :return: url to make pipeline version active
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{version_id}/set-active'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        version_id=version_id)
    return request


def create_workflow_url(config, domain_id):
    """
    returns URL to create workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url for the workflow creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows'.format(ip=config['ip'], port=config['port'],
                                                                                 protocol=config['protocol'],
                                                                                 domain_id=domain_id)
    return request


def configure_workflow_url(config, domain_id, workflow_id):
    """
    returns URL to configure workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param workflow_id: Identifier for workflow
    :type workflow_id: string
    :return: url to configure workflow
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/config-migration'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        workflow_id=workflow_id)
    return request


def trigger_workflow_url(config, domain_id, workflow_id):
    """
    returns URL to trigger a workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param workflow_id: Identifier for workflow
    :type workflow_id: string
    :return: url to trigger workflow
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/start'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        workflow_id=workflow_id)
    return request


def get_workflow_status_url(config, workflow_id, workflow_run_id,domain_id):
    """
    returns URL to fetch status of workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param workflow_id: Identifier for workflow id
    :type workflow_id: string
    :param workflow_run_id: Identifier for workflow run id
    :type workflow_run_id: string
    :return: url to fetch status of workflow using v3 api
    """
    if domain_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/workflows/{workflow_id}/runs/{workflow_run_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            workflow_id=workflow_id,
            workflow_run_id=workflow_run_id)
        return request
    else:
        request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/runs/{workflow_run_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            workflow_id=workflow_id,
            workflow_run_id=workflow_run_id,
            domain_id=domain_id)
        return request

def get_all_workflows_url(config):
    """
    returns URL to fetch all workflows in Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url fetch all workflows in Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_all_workflows_runs_url(config):
    """
    returns URL to fetch all workflows runs in Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url fetch all workflows in Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/workflow_runs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_all_workflow_run_jobs_url(config, run_id):
    """
    returns URL to fetch all workflows run jobs in Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :param run_id: Workflow run id to get the jobs under it
    :type run_id: String
    :return: url fetch all workflows in Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/workflow_runs/{run_id}/jobs'.format(
        run_id=run_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def restart_multiple_workflows_url(config):
    """
    returns URL to restart multiple workflows using Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url restart multiple workflows using Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/workflows/restart'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def cancel_multiple_workflows_url(config):
    """
    returns URL to cancel multiple workflows using Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url cancel multiple workflows using Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/workflows/cancel'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_all_workflows_runs_url_with_domain_id(config, domain_id, workflow_id):
    """
    returns URL to fetch all workflows runs in Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url fetch all workflows in Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/runs'.format(
        domain_id=domain_id,
        workflow_id=workflow_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def create_data_connection(config):
    """
    returns URL to create data connection using v3 api
    :param config: client configurations
    :type config: dict
    :return: url for data connection creation
    """
    request = '{protocol}://{ip}:{port}/v3/admin/data-connections'.format(ip=config['ip'],
                                                                          port=config['port'],
                                                                          protocol=config['protocol']
                                                                          )
    return request


def get_data_connection(config, dataconnection_id=None):
    """
    returns URL to get data connection details using v3 api
    :param config: client configurations
    :type config: dict
    :param dataconnection_id: Identifier for data connection
    :type dataconnection_id: string
    :return: url to get data connection details
    """
    if dataconnection_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/data-connections'.format(
            ip=config['ip'],
            port=config['port'],
            protocol=config['protocol'])
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/data-connections/{dataconnection_id}'.format(
            ip=config['ip'],
            port=config['port'],
            protocol=config['protocol'],
            dataconnection_id=dataconnection_id)
    return request


def get_parent_entity_url(config):
    """
    returns URL to get parent entity
    :param config: client configurations
    :type config: dict
    :return: url to get parent entity
    """
    request = '{protocol}://{ip}:{port}/v3/admin/entity'.format(ip=config['ip'],
                                                                port=config['port'],
                                                                protocol=config['protocol']
                                                                )
    return request


def get_pipeline_details(config, domain_id, pipeline_id):
    """
        returns URL to get pipeline details using v3 api
        :param config: client configurations
        :type config: dict
        :param domain_id: Identifier for domain_id
        :type domain_id: string
        :param pipeline_id: Identifier for pipeline id
        :type pipeline_id: string
        :return: url to get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def get_environment_details(config, env_id=None):
    """
    returns URL to get environment details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :return: url to get environment details
    """
    if env_id is not None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'])
    return request


def get_environment_storage_details(config, env_id, storage_id=None):
    """
    returns URL to get environment storage details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param storage_id: Identifier for storage_id
    :type storage_id: string
    :return: url to get environment storage details
    """
    if storage_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-storage'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-storage/{storage_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            storage_id=storage_id)
    return request


def get_environment_compute_details(config, env_id, compute_id=None):
    """
    returns URL to get environment compute details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param compute_id: Identifier for compute_id
    :type compute_id: string
    :return: url to get environment compute details
    """
    if compute_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-compute-template'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-compute-template/{compute_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            compute_id=compute_id)
    return request


def get_environment_interactive_compute_details(config, env_id, compute_id=None):
    """
    returns URL to get environment interactive compute details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param compute_id: Identifier for compute_id
    :type compute_id: string
    :return: url to get environment interactive compute details
    """
    if compute_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-interactive-clusters'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-interactive-clusters/{compute_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            compute_id=compute_id)
    return request


def submit_source_job(config, source_id):
    """
        returns URL to submit source jobs using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: Identifier for source
        :type source_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        source_id=source_id)
    return request


def resubmit_job_url(config, job_id):
    """
        returns URL to submit source jobs using v3 api
        :param config: client configurations
        :type config: dict
        :param job_id: Identifier for job
        :type job_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/resubmit'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        job_id=job_id)
    return request


def get_advanced_config_url(config, source_id, table_id=None):
    """
        returns URL to list advanced configurations of source using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: source identifier
        :type: source_id: string
        :param table_id: table identifier
        :type: table_id: string
    """
    if table_id is None:
        request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations/advance'.format(ip=config['ip'],
                                                                                                  port=config['port'],
                                                                                                  protocol=config[
                                                                                                      'protocol'],
                                                                                                  source_id=source_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/advance'.format(
            ip=config['ip'],
            port=config['port'],
            protocol=config[
                'protocol'],
            source_id=source_id, table_id=table_id)

    return request


def get_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to get pipeline details using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def delete_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to delete pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url delete pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def update_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to update pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :param pipeline_version_id: Identifier for pipeline version id
    :type pipeline_version_id: string
    :return: url update pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def pipeline_sql_import_url(config, domain_id, pipeline_id):
    """
    returns URL to import sql into pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url import sql into pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/sql-import'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def get_table_configuration(config, source_id, table_id):
    """
    returns URL to configure the get the table configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id:
    :return: url to configure the source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}'.format(ip=config['ip'],
                                                                                         port=config['port'],
                                                                                         protocol=config['protocol'],
                                                                                         source_id=source_id,
                                                                                         table_id=table_id)
    return request


def update_table_ingestion_config_url(config, source_id, table_id):
    """
    returns URL to update the table configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    :return: url to update the table configurations
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/ingestion'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def table_export_config_url(config, source_id, table_id):
    """
    returns URL to get/update the table export configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/export'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def get_ingestion_metrics_source_url(config, source_id):
    """
    returns URL to get ingestion metrics using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/reports/table-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id)
    return request


def get_ingestion_metrics_table_url(config, source_id, table_id):
    """
    returns URL to get ingestion metrics using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/reports/ingestion-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def get_job_metrics_url(config, source_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs/{job_id}/reports/job-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_ingestion_metrics_admin_url(config, job_id):
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/reports/ingestion-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id,
    )
    return request


def get_export_metrics_source_url(config, source_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs/{job_id}/reports/export-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_source_file_paths_url(config, source_id, table_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/jobs/{job_id}/reports/sourceFilesPath'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        table_id=table_id,
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_export_metrics_url(config, job_id):
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/reports/export-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id
    )
    return request


def get_pipeline_build_metrics_url(config, job_id):
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/reports/pipeline-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id
    )
    return request


def get_pipeline_lineage_url(config, domain_id, pipeline_id, pipeline_version_id, column_name, node_id):
    """
        returns URL to get pipeline lineage information using v3 api
        :param config: client configurations
        :type config: dict
        :param domain_id: Identifier for domain_id
        :type domain_id: string
        :param pipeline_id: Identifier for pipeline id
        :type pipeline_id: string
        :param pipeline_version_id: Identifier for pipeline id
        :type pipeline_version_id: string
        :param column_name: Identifier for column name
        :type column_name: string
        :param node_id: Identifier for node id
        :type node_id: string
        :return: url to get pipeline lineage information
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}/lineage?port_type=OUTPUT&column_name={column_name}&node_id={node_id}'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id,
        column_name=column_name,
        node_id=node_id
    )
    return request


def unlock_locked_entities(config):
    """
    returns URL to unlock locked entities
    :param config: client configurations
    :type config: dict
    """
    request = '{protocol}://{ip}:{port}/v3/admin/unlock-entities'.format(ip=config['ip'],
                                                                         port=config['port'],
                                                                         protocol=config['protocol'])
    return request


def trigger_pipeline_build_url(config, domain_id, pipeline_id):
    """
    returns URL to trigger pipeline
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to trigger pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def create_source_extension_url(config):
    """
    returns URL to create source extension
    :param config: client configurations
    :type config: dict
    :return: url to create source extension
    """
    request = '{protocol}://{ip}:{port}/v3/admin/source-extensions'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_jobs_url(config):
    """
    returns URL to get jobs details
    :param config: client configurations
    :type config: dict
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/admin/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_interactive_jobs_url(config, source_id):
    """
    returns URL to get interactive jobs details
    :param config: client configurations
    :type config: dict
    :param source_id: Id of the source
    :type source_id: String
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/interactive-jobs'.format(
        source_id=source_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_admin_jobs_url(config):
    """
    returns URL to get jobs details
    :param config: client configurations
    :type config: dict
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/admin/admin-jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_pipeline_jobs_url(config, domain_id, pipeline_id):
    """
    returns URL to get jobs details
    :param config: client configurations
    :type config: dict
    :param domain_id: Id of the domain for given pipeline
    :type domain_id: String
    :param pipeline_id: Id of the pipeline
    :type pipeline_id: String
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/jobs'.format(
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_cancel_job_url(config, job_id):
    """
    returns URL to get jobs details
    :param config: client configurations
    :type config: dict
    :param job_id: Id of job to cancel
    :type job_id: String
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/jobs/{job_id}/cancel'.format(
        job_id=job_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def list_service_authentication_url(config):
    """
    returns URL to list service authentication using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list service authentication
    """
    request = '{protocol}://{ip}:{port}/v3/admin/manage-secrets/service-auth'.format(ip=config['ip'],
                                                                                     port=config['port'],
                                                                                     protocol=config['protocol'])
    return request


def list_secret_stores_url(config):
    """
    returns URL to list secret stores using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list secret stores
    """
    request = '{protocol}://{ip}:{port}/v3/admin/manage-secrets/secret-store'.format(ip=config['ip'],
                                                                                     port=config['port'],
                                                                                     protocol=config['protocol'])
    return request


def get_query_as_table_url(config, source_id):
    """
        returns URL to create query as table
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/query_tables'.format(
        source_id=source_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_pipeline_group_jobs_base_url(config, domain_id, pipeline_group_id):
    """
        returns URL to get list of pipeline group jobs
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipeline-groups/{pipeline_group_id}/jobs'.format(
        domain_id=domain_id,
        pipeline_group_id=pipeline_group_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_pipeline_group_base_url(config, domain_id):
    """
        returns URL to get details of pipeline group
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipeline-groups/'.format(
        domain_id=domain_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_accessible_pipeline_groups_url(config, domain_id):
    """
    returns URL to get accessible pipeline group
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-pipeline-groups'.format(
        domain_id=domain_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def advance_config_under_pipeline_groups_base_url(config, domain_id, pipeline_group_id):
    """
    returns base URL to get the advance config under pipeline group.
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipeline-groups/{pipeline_group_id}/configurations/advance/'.format(
        domain_id=domain_id, pipeline_group_id=pipeline_group_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_domain_workflow_schedules_url(config, domain_id):
    """
    Returns URL to get Schedules of Workflows belonging to the Domain.
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/schedules'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], domain_id=domain_id
    )
    return request


def get_workflow_schedule_url(config, domain_id, workflow_id):
    """
    Returns URL to get Schedules of particular Workflow belonging to the Domain.
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/schedules'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], domain_id=domain_id, workflow_id=workflow_id
    )
    return request


def get_enable_workflow_schedule_url(config, domain_id, workflow_id):
    """
    Returns URL to Enable Workflow Schedule.
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/schedules/enable'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], domain_id=domain_id, workflow_id=workflow_id
    )
    return request


def get_disable_workflow_schedule_url(config, domain_id, workflow_id):
    """
    Returns URL to Disable Workflow Schedule.
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/schedules/disable'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], domain_id=domain_id, workflow_id=workflow_id
    )
    return request


def update_workflow_schedule_user_url(config, domain_id, workflow_id):
    """
    Return URL to Change Workflow Schedule User
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/schedules/user'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], domain_id=domain_id, workflow_id=workflow_id
    )
    return request


def pause_multiple_workflows_url(config):
    """
    returns url to pause multiple workflows
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/pause'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def resume_multiple_workflows_url(config):
    """
    returns url to resume workflow
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/resume'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_workflow_runs_url(config):
    """
    returns url to get workflow runs
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/runs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_workflow_tasks_url(config):
    """
    returns url to get workflow tasks
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/tasks'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_workflow_run_id_details_url(config, workflow_id, run_id):
    """
    returns url to get workflow run id details
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/{workflow_id}/runs/{run_id}'.format(
        workflow_id=workflow_id, run_id=run_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def source_table_schema_url(config, source_id, table_id):
    """
    returns url to get table schema details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/schema'.format(
        source_id=source_id,
        table_id=table_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_file_mappings_for_json_source_url(config, source_id):
    """
    returns url to get table schema details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/file-mappings'.format(
        source_id=source_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_source_audit_logs_url(config, source_id):
    """
    returns url to get source audit logs
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/audit-logs'.format(
        source_id=source_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_table_audit_logs_url(config, source_id, table_id):
    """
    returns url to get table audit logs
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}//tables/table_id/audit-logs'.format(
        source_id=source_id,
        table_id=table_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_table_group_audit_logs_url(config, source_id, table_group_id):
    """
    returns url to get table group audit logs
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups/table_group_id/audit-logs'.format(
        source_id=source_id,
        table_group_id=table_group_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def accessible_pipeline_extensions_url(config, domain_id):
    """
    returns url to get all the Pipeline Extensions associated with this Domain
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-pipeline-extensions'.format(
        domain_id=domain_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def advanced_config_domain_url(config, domain_id):
    """
    returns url to work with adv config of the domain
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/configurations/advance'.format(
        domain_id=domain_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def advanced_config_pipeline_url(config, domain_id, pipeline_id):
    """
    returns url to work with adv config of the pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/configurations/advance'.format(
        domain_id=domain_id, pipeline_id=pipeline_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def table_advanced_base_url(config, source_id, table_id):
    """
    returns url to work with table level advanced configuration
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/advance'.format(
        source_id=source_id,
        table_id=table_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def table_segmentation_base_url(config, source_id, table_id):
    """
        returns url to work with table segmentation
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/segmentation'.format(
        source_id=source_id,
        table_id=table_id,
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_source_job_url(config, source_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs/{job_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_jobs_prodops_url(config):
    """
    returns URL to get jobs details
    :param config: client configurations
    :type config: dict
    :return: URL to get jobs details
    """
    request = '{protocol}://{ip}:{port}/v3/prodops/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def get_metrics_prodops_url(config):
    request = '{protocol}://{ip}:{port}/v3/prodops/metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
    )


def list_secrets_url(config):
    """
        returns url to get secret details in Infoworks
        """
    request = '{protocol}://{ip}:{port}/v3/admin/manage-secrets/secrets'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def list_replicator_sources_url(config):
    """
    returns URL to list replicator sources using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list sources
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/sources'.format(ip=config['ip'], port=config['port'],
                                                                      protocol=config['protocol'])
    return request


def get_replicator_source_url(config, source_id):
    """
    returns URL to the details of a specific replicator source using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: Identifier of Source
    :type source_id: str
    :return: url to the details of a specific replicator source
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/sources/{source_id}'.format(
        ip=config['ip'], port=config['port'], protocol=config['protocol'], source_id=source_id)
    return request


def create_replicator_source_url(config):
    """
    returns URL to create a replicator source using v3 api
    :param config: client configurations
    :type config: dict
    :return: url for the replicator source creation
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/sources'.format(ip=config['ip'], port=config['port'],
                                                                      protocol=config['protocol'])
    return request


def list_replicator_destinations_url(config):
    """
    returns URL to list replicator destinations using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list destinations
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/destinations'.format(ip=config['ip'], port=config['port'],
                                                                           protocol=config['protocol'])
    return request


def get_replicator_destination_url(config, destination_id):
    """
    returns URL to the details of a specific replicator destination using v3 api
    :param config: client configurations
    :type config: dict
    :param destination_id: Identifier of Destination
    :type destination_id: str
    :return: url to the details of a specific replicator destination
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/destinations/{destination_id}'.format(
        ip=config['ip'], port=config['port'], protocol=config['protocol'], destination_id=destination_id)
    return request


def create_replicator_destination_url(config):
    """
    returns URL to create a replicator destination using v3 api
    :param config: client configurations
    :type config: dict
    :return: url for the replicator destination creation
    """
    request = '{protocol}://{ip}:{port}/v3/replicator/destinations'.format(ip=config['ip'], port=config['port'],
                                                                           protocol=config['protocol'])
    return request


def list_replicator_definitions_url(config, domain_id):
    """
    returns URL to list replicator definitions using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier of Domain
    :type domain_id: String
    :return: url to list definitions
    """

    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id,
                protocol=config['protocol'])
    return request


def get_replicator_definition_url(config, domain_id, definition_id):
    """
    returns URL to get replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier of Domain
    :type domain_id: String
    :param definition_id: Identifier of Definition
    :type definition_id: String
    :return: url to list definitions
    """

    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id, definition_id=definition_id,
                protocol=config['protocol'])
    return request


def create_replicator_definition_url(config, domain_id):
    """
    returns URL to create a replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url for the replicator definition creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions'.format(ip=config['ip'],
                                                                                   port=config['port'],
                                                                                   protocol=config['protocol'],
                                                                                   domain_id=domain_id)
    return request


def list_replicator_domains_url(config):
    """
    returns URL to list replicator domains using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def add_replicator_definition_tables_url(config, domain_id, definition_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param definition_id: Identifier for Replicator definition
    :type definition_id: string
    :return: url to add tables to replicator definition
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}/tables' \
        .format(ip=config['ip'], port=config['port'], protocol=config['protocol'], domain_id=domain_id,
                definition_id=definition_id)

    return request


def submit_replication_metacrawl_job_url(config, domain_id, replicator_source_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param replicator_source_id: Identifier for Source
    :type replicator_source_id: string
    :return: url to submit replication metacrawl job
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/replicator-sources/{replicator_source_id}/jobs' \
        .format(ip=config['ip'], port=config['port'], protocol=config['protocol'], domain_id=domain_id,
                replicator_source_id=replicator_source_id)

    return request


def submit_replication_data_job_url(config, domain_id, definition_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param definition_id: Identifier for Definition
    :type definition_id: string
    :return: url to submit replication data job
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}/jobs' \
        .format(ip=config['ip'], port=config['port'], protocol=config['protocol'], domain_id=domain_id,
                definition_id=definition_id)

    return request


def get_replicator_job_summary_url(config, domain_id, definition_id, job_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param definition_id: Identifier for Definition
    :type definition_id: string
    :param job_id: Identifier of Job
    :type job_id: string
    :return: url to submit replication data job
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}/jobs/{job_id}/summary' \
        .format(ip=config['ip'], port=config['port'], protocol=config['protocol'], domain_id=domain_id,
                definition_id=definition_id, job_id=job_id)
    return request


def add_replicator_sources_to_domain_url(config, domain_id):
    """
    returns URL to add replicator sources to domains

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url to add replicator sources to domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-replicator-sources' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id, protocol=config['protocol'])
    return request


def add_replicator_destinations_to_domain_url(config, domain_id):
    """
    returns URL to add replicator destinations to domains

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url to add replicator destinations to domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-replicator-destinations' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id, protocol=config['protocol'])
    return request


def list_replication_schedules_url(config, domain_id, definition_id):
    """
    returns replication schedules URL

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param definition_id : Identifier for Replicator definition
    :type definition_id: string
    :return: replication schedules URL
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}/schedules' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id, protocol=config['protocol'],
                definition_id=definition_id)
    return request


def get_replication_schedule_url(config, domain_id, definition_id, schedule_id):
    """
    returns replication schedule URL

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param definition_id : Identifier for Replicator definition
    :type definition_id: string
    :param schedule_id : Identifier of schedule
    :type schedule_id: string
    :return: replication schedules URL
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/definitions/{definition_id}/schedules/{schedule_id}' \
        .format(ip=config['ip'], port=config['port'], domain_id=domain_id, protocol=config['protocol'],
                definition_id=definition_id, schedule_id=schedule_id)
    return request


def get_replicator_source_tables_url(config, source_id):
    """
    returns replicator source tables URL

    :param config: client configurations
    :type config: dict
    :param source_id: Identifier for replicator source
    :type source_id: string
    :return: replicator source tables url
    """
    request = '{protocol}://{ip}:{port}/v3//replicator/sources/{source_id}/tables' \
        .format(ip=config['ip'], port=config['port'], source_id=source_id, protocol=config['protocol'])
    return request


def pipeline_version_base_url(config, domain_id, pipeline_id):
    """
    returns URL to work on pipeline versions using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :return: url to work on pipeline version
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions'.format(ip=config['ip'],
                                                                                                        port=config[
                                                                                                            'port'],
                                                                                                        protocol=config[
                                                                                                            'protocol'],
                                                                                                        domain_id=domain_id,
                                                                                                        pipeline_id=pipeline_id)
    return request


def workflow_lineage_url(config, domain_id, workflow_id):
    """
    returns URL for workflow lineage
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param workflow_id: Identifier for workflow
    :type workflow_id: string
    :return: url for workflow lineage
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/lineage'.format(ip=config['ip'],
                                                                                                       port=config[
                                                                                                           'port'],
                                                                                                       protocol=config[
                                                                                                           'protocol'],
                                                                                                       domain_id=domain_id,
                                                                                                       workflow_id=workflow_id)
    return request


def validate_pipeline_url(config, domain_id, pipeline_id, pipeline_version_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}/validate'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id, pipeline_version_id=pipeline_version_id)
    return request


def url_to_refresh_sample_data_of_pipeline(config, domain_id, pipeline_id, pipeline_version_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}/refresh-sample-data'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id, pipeline_version_id=pipeline_version_id)
    return request


def url_to_get_pipeline_node_details(config, domain_id, pipeline_id, pipeline_version_id, node_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}/nodes/{node_id}/'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id, pipeline_version_id=pipeline_version_id, node_id=node_id)
    return request


def url_for_pipeline_version_audits(config, domain_id, pipeline_id, version_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{version_id}/pipeline-version-audits'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id, version_id=version_id)
    return request


def url_for_pipeline_audits(config, domain_id, pipeline_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/pipeline-audits'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id)
    return request


def import_sql_mappings_url(config, domain_id, pipeline_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/import-sql-mappings'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, pipeline_id=pipeline_id)
    return request


def url_to_get_accessible_pipelines(config, domain_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-pipelines'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            protocol=config[
                                                                                                'protocol'],
                                                                                            domain_id=domain_id
                                                                                            )
    return request


def url_to_get_accessible_workflows(config, domain_id):
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-workflows'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            protocol=config[
                                                                                                'protocol'],
                                                                                            domain_id=domain_id
                                                                                            )
    return request


def list_domains_admin_url(config):
    """
    returns Admin URL to list domains using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list domains
    """
    request = '{protocol}://{ip}:{port}/v3/admin/domains'.format(ip=config['ip'], port=config['port'],
                                                                  protocol=config['protocol'])
    return request


def stop_interactive_cluster_url(config):
    request = '{protocol}://{ip}:{port}/v3/admin/environment/cluster/service/stop'.format(ip=config['ip'],
                                                                                          port=config['port'],
                                                                                          protocol=config[
                                                                                              'protocol'],
                                                                                          )
    return request


def terminate_persistent_cluster_aysnc(config):
    request = '{protocol}://{ip}:{port}/v3/admin/environment/cluster/terminate-async'.format(ip=config['ip'],
                                                                                             port=config['port'],
                                                                                             protocol=config[
                                                                                                 'protocol'],
                                                                                             )
    return request


def restart_persistent_cluster_url(config):
    request = '{protocol}://{ip}:{port}/v3/admin/environment/cluster/restart'.format(ip=config['ip'],
                                                                                     port=config['port'],
                                                                                     protocol=config['protocol'],
                                                                                     )
    return request


def stop_persistent_cluster_url(config):
    request = '{protocol}://{ip}:{port}/v3/admin/environment/cluster/stop'.format(ip=config['ip'],
                                                                                  port=config['port'],
                                                                                  protocol=config['protocol'],
                                                                                  )
    return request


def url_to_list_accessible_sources(config, domain_id):
    request = '{protocol}://{ip}:{port}/v3/admin/domains/{domain_id}/accessible-sources-to-add'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id)
    return request


def url_to_list_accessible_pipeline_extensions(config, domain_id):
    request = '{protocol}://{ip}:{port}/v3/admin/domains/{domain_id}/accessible-pipeline-extensions-to-add'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id)
    return request


def url_to_list_accessible_domains(config, domain_id):
    request = '{protocol}://{ip}:{port}/v3/admin/domains/{domain_id}/accessible-domains-to-add'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id)
    return request


def delete_user_schedules_url(config, user_id):
    request = '{protocol}://{ip}:{port}/v3/admin/users/{user_id}/delete-schedules'.format(ip=config['ip'],
                                                                                          port=config['port'],
                                                                                          protocol=config[
                                                                                              'protocol'],
                                                                                          user_id=user_id)
    return request


def list_all_schedules_url(config):
    request = '{protocol}://{ip}:{port}/v3/admin/schedules'.format(ip=config['ip'],
                                                                   port=config['port'],
                                                                   protocol=config['protocol'],
                                                                   )
    return request


def get_crawl_job_summary_url(config, source_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs/{job_id}/logs'.format(ip=config['ip'],
                                                                                          port=config['port'],
                                                                                          protocol=config['protocol'],
                                                                                          source_id=source_id,
                                                                                          job_id=job_id
                                                                                          )
    return request


def get_table_group_schedule_url(config, source_id, table_group_id):
    """
    Returns URL to get Table Group Schedule
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups/{table_group_id}/schedules'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], source_id=source_id,
        table_group_id=table_group_id)
    return request


def get_enable_table_group_schedule_url(config, source_id, table_group_id):
    """
    Returns URL to Enable Table Group Schedule
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups/{table_group_id}/schedules/enable'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], source_id=source_id,
        table_group_id=table_group_id)
    return request


def get_disable_table_group_schedule_url(config, source_id, table_group_id):
    """
    Returns URL to Disable Table Group Schedule
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups/{table_group_id}/schedules/disable'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], source_id=source_id,
        table_group_id=table_group_id)
    return request


def configure_table_group_schedule_user_url(config, source_id, table_group_id):
    """
    Return URL to configure Table Group schedule user
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups/{table_group_id}/schedules/user'.format(
        protocol=config['protocol'], ip=config['ip'], port=config['port'], source_id=source_id,
        table_group_id=table_group_id)
    return request


def get_pipeline_job_summary_url(config, domain_id, pipeline_id, job_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for Pipeline
    :type pipeline_id: string
    :param job_id: Identifier of Job
    :type job_id: string
    :return: url to submit replication data job
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/jobs/{job_id}/summary' \
        .format(ip=config['ip'], port=config['port'], protocol=config['protocol'], domain_id=domain_id,
                pipeline_id=pipeline_id, job_id=job_id)
    return request


def get_pipeline_group_job_summary_url(config, domain_id, pipeline_group_id, job_id):
    """
    returns URL to add tables to replicator definition using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_group_id: Identifier for Pipeline Group
    :type pipeline_group_id: string
    :param job_id: Identifier of Job
    :type job_id: string
    :return: url to submit replication data job
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipeline-groups/{pipeline_group_id}/jobs' \
              '/{job_id}/summary'.format(ip=config['ip'], port=config['port'], protocol=config['protocol'],
                                         domain_id=domain_id, pipeline_group_id=pipeline_group_id, job_id=job_id)
    return request


def stop_streaming_job_url(config, source_id):
    """
        returns URL to stop streaming job
        :param config: client configurations
        :type config: dict
        :param source_id: Identifier for source
        :type source_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/stop-streaming'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        source_id=source_id)
    return request



import os
from uuid import uuid4
import networkx as nx


def get_lineage_dump(edges, entities, entity_type, files_dumped, files_overwrite, config_file_dump_path):
    G = nx.DiGraph()
    G.add_edges_from(edges)
    nodes = [(i, {"filename": files_dumped.get(i, "")}) for i in entities]
    G.add_nodes_from(nodes)
    target_file_path = os.path.join(os.path.join(config_file_dump_path, "modified_files"), f"{entity_type}.csv")
    target_file_path_gexf = os.path.join(os.path.join(config_file_dump_path, "modified_files"), f"{entity_type+'_'+str(uuid4())}.gexf")
    if files_overwrite:
        open(target_file_path, 'w').close()
        mode = "a"
    else:
        mode = "a"
    nx.write_gexf(G, target_file_path_gexf)
    f = open(target_file_path, mode)
    for i in nx.descendants(G, 'root'):  # Returns tree from top to bottom
        file_name = files_dumped.pop(i, None)
        if file_name is not None:
            f.write(file_name)
            f.write("\n")
    f.close()



import copy
import json
import time
import requests
import yaml
import configparser
from infoworks.sdk.url_builder import get_source_details_url,list_secrets_url,create_domain_url, \
    get_environment_interactive_compute_details, restart_persistent_cluster_url
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.source_response import SourceResponse
from infoworks.sdk.local_configurations import Response
from infoworks.sdk.cicd.upload_configurations.update_configurations import InfoworksDynamicAccessNestedDict
from infoworks.sdk.cicd.upload_configurations.local_configurations import PRE_DEFINED_MAPPINGS

class CdataSource:
    def __init__(self):
        self.configuration_obj = None
        self.source_config_path = None
        self.environment_id = None
        self.storage_id = None
        self.secrets = None

    def set_variables(self, environment_id, storage_id, source_config_path, secrets=None, replace_words=""):
        self.storage_id = storage_id
        self.environment_id = environment_id
        self.source_config_path = source_config_path
        with open(self.source_config_path, 'r') as file:
            json_string = file.read()
            if replace_words != "":
                for key, value in [item.split("->") for item in replace_words.split(";")]:
                    json_string = json_string.replace(key, value)
        self.configuration_obj = IWUtils.ejson_deserialize(json_string)
        self.secrets = secrets

    def update_table_schema_and_database(self, type, mappings):
        data = self.configuration_obj
        for table in data.get("configuration", {}).get("table_configs", []):
            if type == "target_schema":
                schema_from_config = table.get("configuration", {}).get("configuration", {}).get("target_schema_name",
                                                                                                 "")
                if schema_from_config != "" and schema_from_config.lower() in mappings.keys():
                    table["configuration"]["configuration"]["target_schema_name"] = mappings.get(
                        schema_from_config.lower())
            elif type == "stage_schema":
                schema_from_config = table.get("configuration", {}).get("configuration", {}).get("staging_schema_name",
                                                                                                 "")
                if schema_from_config != "" and schema_from_config.lower() in mappings.keys():
                    table["configuration"]["configuration"]["staging_schema_name"] = mappings.get(
                        schema_from_config.lower())
            elif type == "database":
                database_from_config = table.get("configuration", {}).get("configuration", {}).get(
                    "target_database_name", "")
                if database_from_config != "" and database_from_config.lower() in mappings.keys():
                    table["configuration"]["configuration"]["target_database_name"] = mappings.get(
                        database_from_config.lower())
            else:
                pass

    def update_mappings_for_configurations(self, mappings):
        config = configparser.ConfigParser()
        config.read_dict(mappings)
        d = InfoworksDynamicAccessNestedDict(self.configuration_obj)
        for section in config.sections():
            if section in PRE_DEFINED_MAPPINGS:
                continue
            print("section:", section)
            try:
                final = d.setval(section.split("$"), dict(config.items(section)))
            except KeyError as e:
                pass
        self.configuration_obj = d.data
        if "configuration$source_configs$data_lake_schema" in config.sections():
            self.update_table_schema_and_database("target_schema",
                                                  dict(config.items("configuration$source_configs$data_lake_schema")))
        if "configuration$source_configs$staging_schema_name" in config.sections():
            self.update_table_schema_and_database("stage_schema", dict(
                config.items("configuration$source_configs$staging_schema_name")))
        if "configuration$source_configs$target_database_name" in config.sections():
            self.update_table_schema_and_database("database", dict(
                config.items("configuration$source_configs$target_database_name")))

    def get_secret_id_from_name(self, cicd_client, secret_name):
        secret_id = None
        get_secret_details_url = list_secrets_url(cicd_client.client_config) + '?filter={"name":"' + secret_name + '"}'
        response = cicd_client.call_api("GET", get_secret_details_url,
                                        IWUtils.get_default_header_for_v3(cicd_client.client_config['bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
            result = parsed_response.get("result", [])

            if len(result) > 0:
                secret_id = result[0]["id"]
                cicd_client.logger.info("Found secret id {} ".format(secret_id))
                return secret_id
            else:
                cicd_client.logger.info("Secret id is {} ".format(None))
                return None

    def poll_cluster_start_status(self, cicd_client, environment_id, compute_id):
        list_of_persistent_cluster_url = get_environment_interactive_compute_details(cicd_client.client_config,
                                                                                     env_id=environment_id,
                                                                                     compute_id=compute_id)
        response = cicd_client.call_api("GET", list_of_persistent_cluster_url,
                                        IWUtils.get_default_header_for_v3(cicd_client.client_config['bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        return parsed_response.get("result", {}).get("cluster_state", "")

    def start_interactive_cluster(self, src_client_obj, environment_id, default_section_mappings):
        retries = 1
        ttl_for_cluster_status_fetch = default_section_mappings.get("ttl_for_cluster_status_fetch", 120)
        ttl_for_cluster_status_fetch = int(ttl_for_cluster_status_fetch) if ttl_for_cluster_status_fetch != "" else 120
        # print("ttl_for_cluster_status_fetch",ttl_for_cluster_status_fetch)
        while retries != 0:
            # added this because it takes about 120 seconds for ttl to expire and refresh the cluster status
            list_of_persistent_cluster_url = get_environment_interactive_compute_details(
                config=src_client_obj.client_config,
                env_id=environment_id) + '?filter={"interactive_job_submit_enabled":true}'
            response = src_client_obj.call_api("GET", list_of_persistent_cluster_url,
                                               IWUtils.get_default_header_for_v3(
                                                   src_client_obj.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(response.content)
            time.sleep(ttl_for_cluster_status_fetch)
            retries = retries - 1
        result = parsed_response.get("result", [])
        print("result", result)
        result = result[0]
        status = result.get("cluster_state", "")
        compute_id = result.get("id", "")
        environment_id = result.get("environment_id", "")
        if status in ["TERMINATED", "STOPPED"]:
            restart_cluster_template_body = \
                {
                    "compute_template_id": compute_id,
                    "type": "persistent"
                }
            start_cluster_url = restart_persistent_cluster_url(config=src_client_obj.client_config)
            response = src_client_obj.call_api("PUT", start_cluster_url,
                                               IWUtils.get_default_header_for_v3(
                                                   src_client_obj.client_config['bearer_token']),
                                               data=restart_cluster_template_body)
            parsed_response = IWUtils.ejson_deserialize(response.content)
            print("Restart persistent cluster response:")
            print(parsed_response)
            if parsed_response.get("result", {}).get("message", "") == "Successfully submitted cluster restart request":
                while (status not in ["RUNNING", "ERROR"]):
                    time.sleep(30)
                    status = self.poll_cluster_start_status(src_client_obj, environment_id=environment_id,
                                                            compute_id=compute_id)
            else:
                print("Something went wrong while restarting the interactive cluster")
            if status == "RUNNING":
                print("Started interactive cluster successfully!")
            else:
                print("Interactive cluster start failed!")
                raise Exception(
                    "Interactive cluster start failed!Please start the cluster manually and retrigger CICD process")
        elif status == "RUNNING":
            print("Interactive Cluster is already running!")
        else:
            print(f"Unknown cluster state {status}")

    def create_cdata_source(self, src_client_obj):
        data = self.configuration_obj["configuration"]["source_configs"]
        create_cdata_source_payload = {
            "name": data["name"],
            "type": data["type"],
            "sub_type": data["sub_type"],
            "data_lake_path": data["data_lake_path"],
            "data_lake_schema": data["data_lake_schema"] if "data_lake_schema" in data else "",
            "environment_id": self.environment_id,
            "storage_id": self.storage_id,
            "is_source_ingested": True
        }
        if data.get("target_database_name", ""):
            create_cdata_source_payload["target_database_name"] = data.get("target_database_name", "")
        if data.get("staging_schema_name", ""):
            create_cdata_source_payload["staging_schema_name"] = data.get("staging_schema_name", "")
        additional_keys_in_source_config = data.keys()
        for key in additional_keys_in_source_config:
            if key not in ["connection"] and key not in create_cdata_source_payload.keys():
                create_cdata_source_payload[key] = data[key]
        # adding associated domains if any
        accessible_domain_names = data.get("associated_domain_names", [])
        accessible_domain_ids = []
        for domain_name in accessible_domain_names:
            domain_response = src_client_obj.call_api("GET",
                                                      create_domain_url(
                                                          src_client_obj.client_config) + "?filter={\"name\":\"" + domain_name + "\"}",
                                                      IWUtils.get_default_header_for_v3(
                                                          src_client_obj.client_config['bearer_token']))
            domain_parsed_response = IWUtils.ejson_deserialize(domain_response.content)
            if domain_response.status_code == 200 and len(domain_parsed_response.get("result", [])) > 0:
                result = domain_parsed_response.get("result", [])
                if len(result) > 0:
                    result = result[0]
                    domain_id = result.get("id", None)
                    if domain_id is not None:
                        accessible_domain_ids.append(domain_id)
            if "associated_domain_names" in create_cdata_source_payload.keys():
                create_cdata_source_payload.pop("associated_domain_names", [])
                self.configuration_obj["configuration"]["source_configs"].pop("associated_domain_names", [])
            if len(accessible_domain_ids) > 0:
                create_cdata_source_payload["associated_domains"] = accessible_domain_ids
                self.configuration_obj["configuration"]["associated_domains"] = accessible_domain_ids
        print("create_rdbms_source_payload:", create_cdata_source_payload)
        src_create_response = src_client_obj.create_source(source_config=create_cdata_source_payload)
        if src_create_response["result"]["status"].upper() == "SUCCESS":
            source_id = src_create_response["result"]["response"]["result"]["id"]
            # added below code to update the source due to IPD-23733
            associated_domains = create_cdata_source_payload.get("associated_domains", [])
            if associated_domains:
                src_client_obj.update_source(source_id=source_id,
                                             update_body={"associated_domains": associated_domains})
            return src_create_response
        else:
            src_client_obj.logger.info('Cant create source {} '.format(data["name"]))
            print('Cant create source {} '.format(data["name"]))
            print(src_create_response)
            src_client_obj.logger.info(f"Getting the existing SourceId with name {data['name']} if exists")
            print(f"Getting the existing SourceId with name {data['name']} if exists")
            filter_condition = IWUtils.ejson_serialize({"name": data['name']})
            source_detail_url = get_source_details_url(
                src_client_obj.client_config) + f"?filter={{filter_condition}}".format(
                filter_condition=filter_condition)
            response = requests.get(source_detail_url,
                                    headers={'Authorization': 'Bearer ' + src_client_obj.client_config['bearer_token'],
                                             'Content-Type': 'application/json'}, verify=False)
            if response.status_code == "406":
                headers = src_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + src_client_obj.client_config['bearer_token'],
                     'Content-Type': 'application/json'})
                response = requests.get(source_detail_url, headers=headers, verify=False)
            response = IWUtils.ejson_deserialize(response.content)
            if not response.get('result', None):
                src_client_obj.logger.error("Failed to make an api call to get source details")
                src_client_obj.logger.info(response)
                print("Failed to make an api call to get source details")
                print(response)
                return SourceResponse.parse_result(status=Response.Status.FAILED, source_id=None, response=response)
            else:
                existing_source_id = response['result'][0]['id']
                src_client_obj.logger.info(
                    f"Source Id with the same Source name {data['name']} : {response['result'][0]['id']}")
                print(f"Source Id with the same Source name {data['name']} : {response['result'][0]['id']}")
                # added below code to update the source due to IPD-23733
                associated_domains = create_cdata_source_payload.get("associated_domains", [])
                if associated_domains:
                    src_client_obj.update_source(source_id=existing_source_id,
                                                 update_body={"associated_domains": associated_domains})
                return SourceResponse.parse_result(status=Response.Status.SUCCESS,
                                                   source_id=response['result'][0]['id'], response=response)

    def configure_cdata_source_connection(self, src_client_obj, source_id, override_config_file=None,
                                          read_passwords_from_secrets=False, env_tag="", secret_type="",
                                          config_ini_path=None, dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        source_configs = self.configuration_obj["configuration"]["source_configs"]
        src_name = str(source_configs["name"])
        connection_object = source_configs["connection"]
        if connection_object.get('connection_mode', '') != '':
            connection_object['connection_mode'] = connection_object['connection_mode'].lower()
        else:
            connection_object['connection_mode'] = 'jdbc'
        if override_config_file is not None:
            with open(override_config_file) as file:
                information = yaml.load(file, Loader=yaml.FullLoader)
            if information["source_details"].get(src_name, None) is not None:
                override_keys = information["source_details"].get(src_name).keys()
                for key in override_keys:
                    connection_object[key] = information["source_details"][src_name][key]
        if connection_object.get("password", {}).get("password_type", "") == "secret_store":
            # for RDBMS passwords in keyvault
            secret_name = connection_object["password"]["secret_name"]
            secret_id = self.get_secret_id_from_name(src_client_obj, secret_name)
            if secret_name:
                connection_object["password"]["secret_id"] = secret_id
                connection_object["password"].pop('secret_name', None)
        # if read_passwords_from_secrets and self.secrets["custom_secrets_read"] is True:
        #     encrypted_key_name = f"{env_tag}-" + src_name
        #     decrypt_value = self.secrets.get(encrypted_key_name, "")
        #     if IWUtils.is_json(decrypt_value):
        #         decrypt_value_dict = json.loads(decrypt_value)
        #         for key in decrypt_value_dict.keys():
        #             connection_object[key] = decrypt_value_dict[key]
        # elif read_passwords_from_secrets and self.secrets["custom_secrets_read"] is False:
        #     encrypted_key_name = f"{env_tag}-" + src_name
        #     decrypt_value = src_client_obj.get_all_secrets(secret_type,keys=encrypted_key_name,ini_config_file_path=config_ini_path)
        #     if len(decrypt_value) > 0 and IWUtils.is_json(decrypt_value[0]):
        #         decrypt_value_dict = json.loads(decrypt_value[0])
        #         for key in decrypt_value_dict.keys():
        #             connection_object[key] = decrypt_value_dict[key]
        response = src_client_obj.configure_source_connection(source_id, connection_object=connection_object)
        if response["result"]["status"].upper() != "SUCCESS":
            src_client_obj.logger.info(f"Failed to configure the source {source_id} connection")
            print(f"Failed to configure the source {source_id} connection")
            src_client_obj.logger.info(response)
            print(response)
            return SourceResponse.parse_result(status=Response.Status.FAILED, source_id=source_id, response=response)
        else:
            src_client_obj.logger.info(response)
            print(response)
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, source_id=source_id, response=response)

    def test_source_connection(self, src_client_obj, source_id, dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        response = src_client_obj.source_test_connection_job_poll(source_id, poll_timeout=300,
                                                                  polling_frequency=15, retries=1)
        return SourceResponse.parse_result(status=Response.Status.SUCCESS, source_id=source_id, response=response)

    def browse_source_tables(self, src_client_obj, source_id, dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        filter_tables_properties = self.configuration_obj["filter_tables_properties"]
        response = src_client_obj.browse_source_tables(source_id, filter_tables_properties=filter_tables_properties,
                                                       poll_timeout=300, polling_frequency=15, retries=1)
        return SourceResponse.parse_result(status=response["result"]["status"].upper(), source_id=source_id,
                                           response=response)

    def add_tables_to_source(self, src_client_obj, source_id, dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        tables_already_added_in_source = src_client_obj.list_tables_in_source(source_id).get("result", {}).get(
            "response", {}).get("result", [])
        tables_already_added_in_source = [table.get("schema_name_at_source",
                                                    table.get("catalog_name",
                                                              "")) + "." + table[
                                              "name"] for table in tables_already_added_in_source]
        tables_list = []
        query_tables_list=[]
        tables = self.configuration_obj["configuration"]["table_configs"]
        if len(tables_already_added_in_source) > 0:
            for table in tables:
                if table["configuration"].get("configuration", {}).get("query","")=="":
                    if table["configuration"].get("schema_name_at_source",
                                                  table["configuration"].get("catalog_name", "")) + "." + \
                            table["configuration"][
                                "name"] not in tables_already_added_in_source:
                        temp = {"table_name": table["configuration"]["name"],
                                "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                                "target_schema_name": table["configuration"]["configuration"]["target_schema_name"]}
                        if table["configuration"].get("catalog_name", "") != "":
                            temp["catalog_name"] = table["configuration"]["catalog_name"]
                        if table["configuration"].get("schema_name_at_source", "") != "":
                            temp["schema_name"] = table["configuration"]["schema_name_at_source"]
                        if table["configuration"].get("configuration", {}).get(
                                "target_database_name", "") != "":
                            temp["target_database_name"] = table["configuration"]["configuration"]["target_database_name"]
                        if table.get("table_type", "") != "":
                            temp["table_type"] = table.get("table_type", "TABLE").upper()
                        tables_list.append(copy.deepcopy(temp))
                        src_client_obj.logger.info(
                            f"Adding table {temp['table_name']} to source {source_id} config payload")
                else:
                    if table["configuration"].get("schema_name_at_source",
                                                  table["configuration"].get("catalog_name", "")) + "." + \
                            table["configuration"][
                                "name"] not in tables_already_added_in_source:
                        temp = {
                                "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                                "target_schema_name": table["configuration"]["configuration"]["target_schema_name"],
                                "is_query_table": True,
                                "query": table["configuration"].get("configuration", {}).get("query","")
                                }
                        if table["configuration"].get("configuration", {}).get("target_database_name",
                                                                                                        "") != "":
                            temp["target_database_name"] = table["configuration"]["configuration"]["target_database_name"]
                        if table.get("table_type", "") != "":
                            temp["table_type"] = table.get("table_type", "TABLE").upper()
                        query_tables_list.append(temp)
        else:
            for table in tables:
                if table["configuration"].get("configuration", {}).get("query", "") == "":
                    temp = {"table_name": table["configuration"]["name"],
                            "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                            "target_schema_name": table["configuration"]["configuration"]["target_schema_name"]}
                    if table["configuration"].get("catalog_name", "") != "":
                        temp["catalog_name"] = table["configuration"]["catalog_name"]
                    if table["configuration"].get("schema_name_at_source", "") != "":
                        temp["schema_name"] = table["configuration"]["schema_name_at_source"]
                    if table["configuration"].get("configuration", {}).get("target_database_name",
                                                                                                    "") != "":
                        temp["target_database_name"] = table["configuration"]["configuration"]["target_database_name"]
                    if table.get("table_type", "") != "":
                        temp["table_type"] = table.get("table_type", "TABLE").upper()
                    tables_list.append(copy.deepcopy(temp))
                    src_client_obj.logger.info(f"Adding table {temp['table_name']} to source {source_id} config payload")
                else:
                    temp = {
                            "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                            "target_schema_name": table["configuration"]["configuration"]["target_schema_name"],
                            "is_query_table": True,
                            "query": table["configuration"].get("configuration", {}).get("query", "")
                            }
                    if table["configuration"].get("configuration", {}).get("target_database_name",
                                                                                                    "") != "":
                        temp["target_database_name"] = table["configuration"]["configuration"]["target_database_name"]
                    if table.get("table_type", "") != "":
                        temp["table_type"] = table.get("table_type", "TABLE").upper()
                    query_tables_list.append(temp)
        print("query_tables_list",query_tables_list)
        if len(query_tables_list)>0:
            response = src_client_obj.add_query_tables_to_source(source_id, query_tables_list)
            print("Add query tables API response:", response)
        else:
            src_client_obj.logger.info(f"No new query tables found to add.")
            print(f"No new query tables found to add.")
        print("tables_list",tables_list)
        if len(tables_list) > 0:
            response = src_client_obj.add_tables_to_source(source_id, tables_list)
            print("Add tables API response:",response)
            return SourceResponse.parse_result(status=response["result"]["status"].upper(), source_id=source_id,
                                               response=response)
        else:
            src_client_obj.logger.info(f"No new tables found to add.")
            print(f"No new tables found to add.")
            return SourceResponse.parse_result(status="SUCCESS", source_id=source_id,
                                               response={})

    def update_schema_for_tables(self, src_client_obj, source_id, export_configuration_file=None,
                                 export_config_lookup=True, mappings=None, read_passwords_from_secrets=False,
                                 env_tag="", secret_type="", dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        tables = self.configuration_obj["configuration"]["table_configs"]
        table_schema_update_dict = {}
        for table in tables:
            table_name = table["configuration"]["name"]
            src_client_obj.logger.info(f"Updating the schema information for table {table_name}")
            columns = table["configuration"]["columns"]
            table_update_payload = {"name": table_name, "source": source_id, "columns": columns}
            if table["configuration"].get("schema_name_at_source", "") != "":
                table_document = src_client_obj.list_tables_in_source(source_id, params={
                    "filter": {"origTableName": table["configuration"]["name"],
                               "schemaNameAtSource": table["configuration"]["schema_name_at_source"]}}).get("result",
                                                                                                            {}).get(
                    "response", {}).get("result", [])
            elif table["configuration"].get("catalog_name", ""):
                table_document = src_client_obj.list_tables_in_source(source_id, params={
                    "filter": {"origTableName": table["configuration"]["name"],
                               "catalog_name": table["configuration"]["catalog_name"]}}).get("result", {}).get(
                    "response", {}).get("result", [])
            elif table["configuration"]["configuration"].get("query", ""):
                table_document = src_client_obj.list_tables_in_source(source_id, params={
                    "filter": {"table": table_name}}).get("result", {}).get(
                    "response", {}).get("result", [])
            else:
                print(f"Skipping schema updation for table {table_name} as it did not match any existing table.")
            table_id = None
            if len(table_document) > 0:
                table_document = table_document[0]
                table_id = table_document["id"]
            if table_id is not None:
                response = src_client_obj.update_table_configuration(source_id=source_id, table_id=table_id,
                                                                     config_body=table_update_payload)
                if response["result"]["status"].upper() != "SUCCESS":
                    src_client_obj.logger.error(
                        "Failed to update schema for table {table_name}".format(table_name=table_name))
                    src_client_obj.logger.error(response.get("result", {}).get("response", {}).get("message", ""))
                    table_schema_update_dict[table_name] = (
                        "FAILED", response.get("result", {}).get("response", {}).get("message", ""))
                else:
                    src_client_obj.logger.info(
                        "Successfully updated schema for table {table_name}".format(table_name=table_name))
                    table_schema_update_dict[table_name] = ("SUCCESS", "")
        failed_schema_update_tables = [(table_name, status[1]) for table_name, status in
                                       table_schema_update_dict.items() if status[0].upper() == "FAILED"]
        overall_update_status = "FAILED" if len(failed_schema_update_tables) > 0 else "SUCCESS"
        if overall_update_status == "FAILED":
            response = {f"Tables schema update failed for tables:{failed_schema_update_tables}"}
        else:
            response = {f"Tables schema updated successfully"}
        return SourceResponse.parse_result(status=overall_update_status, source_id=source_id,
                                           response=response)

    def configure_tables_and_tablegroups(self, src_client_obj, source_id, export_configuration_file=None,
                                         export_config_lookup=True, mappings=None, read_passwords_from_secrets=False,
                                         env_tag="", secret_type="", dont_skip_step=True):
        if not dont_skip_step:
            return SourceResponse.parse_result(status="SKIPPED", source_id=source_id)
        if mappings is None:
            mappings = {}
        source_configs = self.configuration_obj["configuration"]["source_configs"]
        src_name = str(source_configs["name"])
        # Update the service account json file mappings if any
        if export_config_lookup and (export_configuration_file is not None or read_passwords_from_secrets):
            for table in self.configuration_obj["configuration"]["table_configs"]:
                # Check if there are any export configurations and passwords to replace
                if table.get("configuration", {}).get("export_configuration", None) is not None:
                    export_configs = table.get("configuration", {}).get("export_configuration")
                    target_type = export_configs.get("target_type", "").upper()
                    table_name = table["configuration"]["name"].upper()
                    override_keys = []
                    if export_configuration_file is not None:
                        with open(export_configuration_file) as file:
                            information = yaml.load(file, Loader=yaml.FullLoader)
                        if information["src_export_details"].get(src_name + "_" + table_name, None) is not None:
                            info_key = src_name + "_" + table_name
                            override_keys = information["src_export_details"].get(src_name + "_" + table_name,
                                                                                  {}).keys()
                        else:
                            info_key = src_name
                            override_keys = information["src_export_details"].get(src_name, {}).keys()
                        for key in override_keys:
                            table["configuration"]["export_configuration"]["connection"][key] = \
                                information["src_export_details"][info_key][key]
                    if target_type.upper() in ["SNOWFLAKE", "POSTGRES"]:
                        pass

                    elif target_type.upper() == "BIGQUERY":
                        if "server_path" not in override_keys:
                            server_path = table["configuration"]["export_configuration"].get("connection", {}).get(
                                "server_path", "")
                            if "gcp_details" in mappings:
                                server_path = mappings["gcp_details"].get("service_json_path")
                            if "service_json_mappings" in mappings:
                                server_path = mappings["service_json_mappings"].get(
                                    server_path.split("/")[-1],
                                    server_path)
                            table["configuration"]["export_configuration"]["connection"][
                                "server_path"] = server_path if server_path != "" else table["configuration"][
                                "export_configuration"].get("connection", {}).get(
                                "server_path", "")
                        table["configuration"]["export_configuration"]["connection"][
                            "upload_option"] = "serverLocation"

        response = src_client_obj.configure_tables_and_tablegroups(source_id, configuration_obj=self.configuration_obj[
            "configuration"])
        if response["result"]["status"].upper() != "SUCCESS":
            src_client_obj.logger.error("Failed to import the source {} (source config path : {})"
                                        .format(source_id, self.source_config_path))
            src_client_obj.logger.error(response.get("message", "") + "(source config path : {})"
                                        .format(self.source_config_path))
            return SourceResponse.parse_result(status="FAILED", source_id=source_id,
                                               response=response)

        else:
            src_client_obj.logger.info(f"Successfully imported source configurations to {source_id}")
            return SourceResponse.parse_result(status=response["result"]["status"].upper(), source_id=source_id,
                                               response=response)



from base64 import b64decode, b64encode
import calendar
from collections import OrderedDict
from datetime import date, datetime
import json
import six

EJSON_KEYWORDS = ("$date", "$type", "$value", "$escape", "$binary")

try:
    from datetime import timezone
except ImportError:
    from datetime import tzinfo, timedelta


    class timezone(tzinfo):
        def __init__(self, utcoffset, name=None):
            self._utcoffset = utcoffset
            self._name = name

        def utcoffset(self, dt):
            return self._utcoffset

        def tzname(self, dt):
            return self._name

        def dst(self, dt):
            return timedelta(0)


    timezone.utc = timezone(timedelta(0), 'UTC')


class UnknownTypeError(ValueError):
    pass


class EJSONEncoder(json.JSONEncoder):

    def __init__(self, *args, **kwargs):
        self.custom_type_hooks = kwargs.pop('custom_type_hooks', ())
        json.JSONEncoder.__init__(self, *args, **kwargs)

    def encode(self, o):
        if self.check_circular:
            markers = {}
        else:
            markers = None

        def _encode(o):
            if isinstance(o, (list, tuple, dict)):
                if markers is not None:
                    marker_id = id(o)
                    if marker_id in markers:
                        raise ValueError("Circular reference detected")
                    markers[marker_id] = o
                try:
                    if isinstance(o, dict):
                        dict_cls = OrderedDict if isinstance(o, OrderedDict) else dict
                        if any(kw in o for kw in EJSON_KEYWORDS):
                            return {"$escape": dict_cls((k, _encode(v)) for k, v in o.items())}
                        return dict_cls((k, _encode(v)) for k, v in o.items())
                    else:
                        return [_encode(v) for v in o]
                finally:
                    if markers is not None:
                        del markers[marker_id]

            if isinstance(o, date):
                return {"$date": int(calendar.timegm(o.timetuple()) * 1000)}

            if isinstance(o, datetime):
                return {"$date": int(calendar.timegm(o.utctimetuple()) * 1000)}

            if six.PY3 and isinstance(o, six.binary_type):
                return {"$binary": b64encode(o).decode()}

            for dict_cls, name, f in self.custom_type_hooks:
                if isinstance(o, dict_cls):
                    return {"$type": name, "$value": f(o)}

            return o

        return json.JSONEncoder.encode(self, _encode(o))


class EJSONDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        self.custom_type_hooks = dict(kwargs.pop('custom_type_hooks', ()))
        json.JSONDecoder.__init__(self, *args, **kwargs)

    def _decode_escaped(self, o):
        if isinstance(o, dict):
            if self.object_pairs_hook is not None:
                return self.object_pairs_hook((k, self._decode(v)) for k, v in o.items())
            return {k: self._decode(v) for k, v in o.items()}
        return self._decode(o)

    def _decode(self, o):
        if isinstance(o, dict):
            if len(o) == 1:
                if "$escape" in o:
                    return self._decode_escaped(o['$escape'])
                if "$date" in o:
                    return self._decode_escaped(o['$date'])
                if "$binary" in o:
                    return b64decode(o['$binary'])
            if len(o) == 2 and "$type" in o and "$value" in o:
                try:
                    reviver = self.custom_type_hooks[o['$type']]
                except KeyError:
                    raise UnknownTypeError(o["$type"])

                return reviver(o["$value"])

            if self.object_pairs_hook is not None:
                return self.object_pairs_hook((k, self._decode(v)) for k, v in o.items())
            return {k: self._decode(v) for k, v in o.items()}

        if isinstance(o, (list, tuple)):
            return [self._decode(v) for v in o]

        return o

    def decode(self, s, *args, **kwargs):
        o = json.JSONDecoder.decode(self, s, *args, **kwargs)
        return self._decode(o)


def dumps(obj, *args, **kwargs):
    return json.dumps(obj, *args, cls=kwargs.pop('cls', EJSONEncoder), **kwargs)


def loads(obj, *args, **kwargs):
    return json.loads(obj, *args, cls=kwargs.pop('cls', EJSONDecoder), **kwargs)
